#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Gen AI operators."""

from __future__ import annotations

import enum
import os.path
import time
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from google.genai.errors import ClientError

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.gen_ai import (
    GenAIGeminiAPIHook,
    GenAIGenerativeModelHook,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from google.genai.types import (
        ContentListUnion,
        ContentListUnionDict,
        CountTokensConfigOrDict,
        CreateBatchJobConfig,
        CreateCachedContentConfigOrDict,
        CreateTuningJobConfigOrDict,
        EmbedContentConfigOrDict,
        GenerateContentConfig,
        ListBatchJobsConfig,
        TuningDatasetOrDict,
    )

    from airflow.providers.common.compat.sdk import Context


class GenAIGenerateEmbeddingsOperator(GoogleCloudBaseOperator):
    """
    Uses the Gemini AI Embeddings API to generate embeddings for words, phrases, sentences, and code.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param model: Required. The name of the model to use for content generation,
        which can be a text-only or multimodal model. For example, `gemini-pro` or
        `gemini-pro-vision`.
    :param contents: Optional. The contents to use for embedding.
    :param config: Optional. Configuration for embeddings.
    :param gcp_conn_id: Optional. The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "contents", "model", "config")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        model: str,
        contents: ContentListUnion | ContentListUnionDict | list[str],
        config: EmbedContentConfigOrDict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.config = config
        self.model = model
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Generating text embeddings...")
        response = self.hook.embed_content(
            project_id=self.project_id,
            location=self.location,
            contents=self.contents,
            model=self.model,
            config=self.config,
        )

        self.log.info("Model response: %s", response)
        context["ti"].xcom_push(key="model_response", value=response)

        return response


class GenAIGenerateContentOperator(GoogleCloudBaseOperator):
    """
    Generate a model response based on given configuration. Input capabilities differ between models, including tuned models.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param model: Required. The name of the model to use for content generation,
        which can be a text-only or multimodal model. For example, `gemini-pro` or
        `gemini-pro-vision`.
    :param contents: Required. The multi-part content of a message that a user or a program
        gives to the generative model, in order to elicit a specific response.
    :param generation_config: Optional. Generation configuration settings.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "generation_config",
        "location",
        "project_id",
        "impersonation_chain",
        "contents",
        "model",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        contents: ContentListUnionDict,
        model: str,
        generation_config: GenerateContentConfig | dict[str, Any] | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.generation_config = generation_config
        self.model = model
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.generate_content(
            project_id=self.project_id,
            location=self.location,
            model=self.model,
            contents=self.contents,
            generation_config=self.generation_config,
        )

        self.log.info("Created Content: %s", response)
        context["ti"].xcom_push(key="model_response", value=response)

        return response


class GenAISupervisedFineTuningTrainOperator(GoogleCloudBaseOperator):
    """
    Create a tuning job to adapt model behavior with a labeled dataset.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param source_model: Required. A pre-trained model optimized for performing natural
        language tasks such as classification, summarization, extraction, content
        creation, and ideation.
    :param training_dataset: Required. Cloud Storage URI of your training dataset. The dataset
        must be formatted as a JSONL file. For best results, provide at least 100 to 500 examples.
    :param tuning_job_config: Optional. Configuration of the Tuning job to be created.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "training_dataset",
        "tuning_job_config",
        "source_model",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        source_model: str,
        training_dataset: TuningDatasetOrDict,
        tuning_job_config: CreateTuningJobConfigOrDict | dict[str, Any] | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.source_model = source_model
        self.training_dataset = training_dataset
        self.tuning_job_config = tuning_job_config
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.supervised_fine_tuning_train(
            project_id=self.project_id,
            location=self.location,
            source_model=self.source_model,
            training_dataset=self.training_dataset,
            tuning_job_config=self.tuning_job_config,
        )

        self.log.info("Tuned Model Name: %s", response.tuned_model.model)  # type: ignore[union-attr,arg-type]
        self.log.info("Tuned Model EndpointName: %s", response.tuned_model.endpoint)  # type: ignore[union-attr,arg-type]

        context["ti"].xcom_push(key="tuned_model_name", value=response.tuned_model.model)  # type: ignore[union-attr,arg-type]
        context["ti"].xcom_push(key="tuned_model_endpoint_name", value=response.tuned_model.endpoint)  # type: ignore[union-attr,arg-type]

        result = {
            "tuned_model_name": response.tuned_model.model,  # type: ignore[union-attr,arg-type]
            "tuned_model_endpoint_name": response.tuned_model.endpoint,  # type: ignore[union-attr,arg-type]
        }

        return result


class GenAICountTokensOperator(GoogleCloudBaseOperator):
    """
    Use Count Tokens API to calculate the number of input tokens before sending a request to Gemini API.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param contents: Required. The multi-part content of a message that a user or a program
        gives to the generative model, in order to elicit a specific response.
    :param model: Required. Model, supporting prompts with text-only input,
        including natural language tasks, multi-turn text and code chat,
        and code generation. It can output text and code.
    :param config: Optional. Configuration for Count Tokens.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "contents", "model", "config")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        contents: ContentListUnion | ContentListUnionDict,
        model: str,
        config: CountTokensConfigOrDict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.model = model
        self.config = config
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.count_tokens(
            project_id=self.project_id,
            location=self.location,
            contents=self.contents,
            model=self.model,
            config=self.config,
        )

        self.log.info("Total tokens: %s", response.total_tokens)
        context["ti"].xcom_push(key="total_tokens", value=response.total_tokens)


class GenAICreateCachedContentOperator(GoogleCloudBaseOperator):
    """
    Create CachedContent resource to reduce the cost of requests that contain repeat content with high input token counts.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for cached content.
    :param cached_content_config: Optional. Configuration of the Cached Content.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "model", "cached_content_config")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        model: str,
        cached_content_config: CreateCachedContentConfigOrDict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.model = model
        self.cached_content_config = cached_content_config
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        cached_content_name = self.hook.create_cached_content(
            project_id=self.project_id,
            location=self.location,
            model=self.model,
            cached_content_config=self.cached_content_config,
        )

        self.log.info("Cached Content Name: %s", cached_content_name)
        context["ti"].xcom_push(key="cached_content", value=cached_content_name)

        return cached_content_name


class BatchJobStatus(enum.Enum):
    """Possible states of batch job in Gemini Batch API."""

    SUCCEEDED = "JOB_STATE_SUCCEEDED"
    PENDING = "JOB_STATE_PENDING"
    FAILED = "JOB_STATE_FAILED"
    RUNNING = "JOB_STATE_RUNNING"
    CANCELLED = "JOB_STATE_CANCELLED"
    EXPIRED = "JOB_STATE_EXPIRED"


class GenAIGeminiCreateBatchJobOperator(GoogleCloudBaseOperator):
    """
    Create Batch job using Gemini Batch API. Use to generate model response for several requests.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param input_source: Required. Source of requests, could be inline requests or file name.
    :param results_folder: Optional. Path to a folder on local machine where file with results will be saved.
    :param create_batch_job_config: Optional. Config for batch job creation.
    :param wait_until_complete: Optional. Await job completion.
    :param retrieve_result: Optional. Push the result to XCom. If the input_source is inline, this pushes
        the execution result. If a file name is specified, this pushes the output file path.
    :param polling_interval: Optional. The interval, in seconds, to poll the job status.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "model",
        "create_batch_job_config",
        "gemini_api_key",
        "input_source",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        model: str,
        input_source: list | str,
        gemini_api_key: str,
        create_batch_job_config: CreateBatchJobConfig | dict | None = None,
        results_folder: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        retrieve_result: bool = False,
        wait_until_complete: bool = False,
        polling_interval: int = 30,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.model = model
        self.input_source = input_source
        self.create_batch_job_config = create_batch_job_config
        self.gemini_api_key = gemini_api_key
        self.retrieve_result = retrieve_result
        self.wait_until_complete = wait_until_complete
        self.polling_interval = polling_interval
        self.results_folder = results_folder

        if self.retrieve_result and not self.wait_until_complete:
            raise AirflowException("Retrieving results is possible only if wait_until_complete set to True")
        if self.results_folder and not isinstance(self.input_source, str):
            raise AirflowException("results_folder works only when input_source is file name")
        if self.results_folder and not os.path.exists(os.path.abspath(self.results_folder)):
            raise AirflowException("path to results_folder does not exist, please provide correct path")

    def _wait_until_complete(self, job, polling_interval: int = 30):
        try:
            while True:
                job = self.hook.get_batch_job(job_name=job.name)
                if job.state.name == BatchJobStatus.SUCCEEDED.value:
                    self.log.info("Job execution completed")
                    break
                if job.state.name in [
                    BatchJobStatus.FAILED.value,
                    BatchJobStatus.EXPIRED.value,
                    BatchJobStatus.CANCELLED.value,
                ]:
                    self.log.error("Job execution was not completed!")
                    break
                self.log.info(
                    "Waiting for job execution, polling interval: %s seconds, current state: %s",
                    self.polling_interval,
                    job.state.name,
                )
                time.sleep(polling_interval)
        except Exception:
            raise AirflowException("Something went wrong during waiting of the batch job.")
        return job

    def _prepare_results_for_xcom(self, job):
        results = []
        if job.dest and job.dest.inlined_responses:
            self.log.info("Results are inline")
            for inline_response in job.dest.inlined_responses:
                if inline_response.response:
                    # Accessing response, structure may vary.
                    try:
                        results.append(inline_response.response.text)
                    except AttributeError:
                        results.append(inline_response.response)
                elif inline_response.error:
                    self.log.warning("Error found in the inline result")
                    results.append(inline_response.error)
        elif job.dest and job.dest.file_name:
            file_content_bytes = self.hook.download_file(file_name=job.dest.file_name)
            file_content = file_content_bytes.decode("utf-8")
            file_name = job.display_name or job.name.replace("/", "-")
            path_to_file = os.path.abspath(f"{self.results_folder}/{file_name}.jsonl")
            with open(path_to_file, "w") as file_with_results:
                file_with_results.writelines(file_content.splitlines(True))
            results = path_to_file

        return results

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        try:
            job = self.hook.create_batch_job(
                model=self.model,
                source=self.input_source,
                create_batch_job_config=self.create_batch_job_config,
            )
        except Exception as e:
            raise AirflowException("Something went wrong during creation of the batch job: %s", e)

        self.log.info("Job with name %s was successfully created!", job.name)
        context["ti"].xcom_push(key="job_name", value=job.name)

        if self.wait_until_complete:
            job = self._wait_until_complete(job, self.polling_interval)
            if self.retrieve_result and job.error is None:
                job_results = self._prepare_results_for_xcom(job)
                context["ti"].xcom_push(key="job_results", value=job_results)

        return dict(job)


class GenAIGeminiGetBatchJobOperator(GoogleCloudBaseOperator):
    """
    Get Batch job using Gemini API.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param job_name: Required. Name of the batch job.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "job_name", "gemini_api_key")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        job_name: str,
        gemini_api_key: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.job_name = job_name
        self.gemini_api_key = gemini_api_key

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        try:
            job = self.hook.get_batch_job(job_name=self.job_name)
        except ValueError:
            raise AirflowException("Job with name %s not found", self.job_name)

        context["ti"].xcom_push(key="job_status", value=job.state)
        return dict(job)


class GenAIGeminiListBatchJobsOperator(GoogleCloudBaseOperator):
    """
    Get list of Batch jobs metadata using Gemini API.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "list_batch_jobs_config",
        "gemini_api_key",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        gemini_api_key: str,
        list_batch_jobs_config: ListBatchJobsConfig | dict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.list_batch_jobs_config = list_batch_jobs_config
        self.gemini_api_key = gemini_api_key

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        jobs_list = self.hook.list_batch_jobs(list_batch_jobs_config=self.list_batch_jobs_config)

        job_names = []
        job_objs = []

        try:
            for job in jobs_list:
                job_names.append(job.name)
                job_objs.append(job.model_dump(exclude={"dest"}))
        except RuntimeError:
            self.log.info("%s jobs found", len(job_names))

        context["ti"].xcom_push(key="job_names", value=job_names)

        return job_objs


class GenAIGeminiDeleteBatchJobOperator(GoogleCloudBaseOperator):
    """
    Queue a batch job for deletion using the Gemini API.

    The job will not be deleted immediately. After submitting it for deletion, it will still be available
    through GenAIGeminiListBatchJobsOperator or GenAIGeminiGetBatchJobOperator for some time.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param job_name: Required. Name of the batch job.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "job_name", "gemini_api_key")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        job_name: str,
        gemini_api_key: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.job_name = job_name
        self.gemini_api_key = gemini_api_key

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        try:
            delete_response = self.hook.delete_batch_job(job_name=self.job_name)
        except ValueError:
            raise AirflowException("Job with name %s was not found", self.job_name)

        self.log.info("Job with name %s was submitted for deletion.", self.job_name)

        if delete_response.error:
            raise AirflowException(
                "Job with name %s was not deleted due to error: %s", self.job_name, delete_response.error
            )

        return delete_response.model_dump()


class GenAIGeminiCancelBatchJobOperator(GoogleCloudBaseOperator):
    """
    Cancel Batch job using Gemini API.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param job_name: Required. Name of the batch job.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "job_name", "gemini_api_key")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        job_name: str,
        gemini_api_key: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.job_name = job_name
        self.gemini_api_key = gemini_api_key

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )
        self.log.info("Cancelling job with name %s ...", self.job_name)

        try:
            self.hook.cancel_batch_job(job_name=self.job_name)
        except ValueError:
            raise AirflowException("Job with name %s was not found", self.job_name)

        self.log.info("Job with name %s was successfully cancelled", self.job_name)


class GenAIGeminiCreateEmbeddingsBatchJobOperator(GoogleCloudBaseOperator):
    """
    Create embeddings Batch job using Gemini Batch API.

    Use to generate embeddings for words, phrases, sentences, and code for several requests.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param input_source: Required. Source of requests, could be inline requests or file name.
    :param results_folder: Optional. Path to a folder on local machine where file with results will be saved.
    :param create_embeddings_config: Optional. Config for batch job creation.
    :param wait_until_complete: Optional. Await job completion.
    :param retrieve_result: Optional. Push the result to XCom. If the input_source is inline, this pushes
        the execution result. If a file name is specified, this pushes the output file path.
    :param polling_interval: Optional. The interval, in seconds, to poll the job status.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "model",
        "create_embeddings_config",
        "gemini_api_key",
        "input_source",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        model: str,
        gemini_api_key: str,
        input_source: dict | str,
        results_folder: str | None = None,
        create_embeddings_config: CreateBatchJobConfig | dict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        wait_until_complete: bool = False,
        retrieve_result: bool = False,
        polling_interval: int = 30,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.model = model
        self.input_source = input_source
        self.create_embeddings_config = create_embeddings_config
        self.gemini_api_key = gemini_api_key
        self.wait_until_complete = wait_until_complete
        self.retrieve_result = retrieve_result
        self.polling_interval = polling_interval
        self.results_folder = results_folder

        if self.retrieve_result and not self.wait_until_complete:
            raise AirflowException("Retrieving results is possible only if wait_until_complete set to True")
        if self.results_folder and not isinstance(self.input_source, str):
            raise AirflowException("results_folder works only when input_source is file name")
        if self.results_folder and not os.path.exists(os.path.abspath(self.results_folder)):
            raise AirflowException("path to results_folder does not exist, please provide correct path")

    def _wait_until_complete(self, job, polling_interval: int = 30):
        try:
            while True:
                job = self.hook.get_batch_job(job_name=job.name)
                if job.state.name == BatchJobStatus.SUCCEEDED.value:
                    self.log.info("Job execution completed")
                    break
                if job.state.name in [
                    BatchJobStatus.FAILED.value,
                    BatchJobStatus.EXPIRED.value,
                    BatchJobStatus.CANCELLED.value,
                ]:
                    self.log.error("Job execution was not completed!")
                    break
                self.log.info(
                    "Waiting for job execution, polling interval: %s seconds, current state: %s",
                    self.polling_interval,
                    job.state.name,
                )
                time.sleep(polling_interval)
        except Exception as e:
            raise AirflowException("Something went wrong during waiting of the batch job: %s", e)
        return job

    def _prepare_results_for_xcom(self, job):
        results = []
        if job.dest and job.dest.inlined_embed_content_responses:
            self.log.info("Results are inline")
            for inline_embed_response in job.dest.inlined_embed_content_responses:
                if inline_embed_response.response:
                    # Accessing response, structure may vary.
                    try:
                        results.append(dict(inline_embed_response.response.embedding))
                    except AttributeError:
                        results.append(inline_embed_response.response)
                elif inline_embed_response.error:
                    self.log.warning("Error found in the inline result")
                    results.append(inline_embed_response.error)
        elif job.dest and job.dest.file_name:
            file_content_bytes = self.hook.download_file(file_name=job.dest.file_name)
            file_content = file_content_bytes.decode("utf-8")
            file_name = job.display_name or job.name.replace("/", "-")
            path_to_file = os.path.abspath(f"{self.results_folder}/{file_name}.jsonl")
            with open(path_to_file, "w") as file_with_results:
                file_with_results.writelines(file_content.splitlines(True))
            results = path_to_file

        return results

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        try:
            embeddings_job = self.hook.create_embeddings(
                model=self.model,
                source=self.input_source,
                create_embeddings_config=self.create_embeddings_config,
            )
        except Exception:
            raise AirflowException("Something went wrong during creation of the embeddings job.")

        self.log.info("Embeddings Job with name %s was successfully created!", embeddings_job.name)
        context["ti"].xcom_push(key="job_name", value=embeddings_job.name)

        if self.wait_until_complete:
            embeddings_job = self._wait_until_complete(embeddings_job, self.polling_interval)
            if self.retrieve_result and embeddings_job.error is None:
                job_results = self._prepare_results_for_xcom(embeddings_job)
                context["ti"].xcom_push(key="job_results", value=job_results)

        return embeddings_job.model_dump()


class GenAIGeminiUploadFileOperator(GoogleCloudBaseOperator):
    """
    Get file uploaded to Gemini Files API.

    The Files API lets you store up to 20GB of files per project, with each file not exceeding 2GB in size.
    Supported types are audio files, images, videos, documents, and others. Files are stored for 48 hours.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param file_path: Required. Path to file on your local machine.
    :param upload_file_config: Optional. Metadata configuration for file upload.
        Defaults to display name and mime type parsed from file_path.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "file_path",
        "gemini_api_key",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        file_path: str,
        gemini_api_key: str,
        upload_file_config: dict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.gemini_api_key = gemini_api_key
        self.file_path = file_path
        self.upload_file_config = upload_file_config

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        try:
            file = self.hook.upload_file(
                path_to_file=self.file_path, upload_file_config=self.upload_file_config
            )
        except RuntimeError as exc:
            raise exc
        except ValueError:
            raise AirflowException("Error during file upload! Check file name or mime type!")
        except FileNotFoundError:
            raise AirflowException("Provided file was not found!")

        self.log.info("File with name %s successfully uploaded!", file.name)
        context["ti"].xcom_push(key="file_name", value=file.name)

        return file.model_dump()


class GenAIGeminiGetFileOperator(GoogleCloudBaseOperator):
    """
    Get file's metadata uploaded to Gemini Files API by using GenAIGeminiUploadFileOperator.

    The Files API lets you store up to 20GB of files per project, with each file not exceeding 2GB in size.
    Files are stored for 48 hours.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param file_name: Required. File name in Gemini Files API to get
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "file_name",
        "gemini_api_key",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        file_name: str,
        gemini_api_key: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.gemini_api_key = gemini_api_key
        self.file_name = file_name

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )
        self.log.info("Looking for file with name: %s", self.file_name)

        try:
            file = self.hook.get_file(file_name=self.file_name)
        except ClientError:
            raise AirflowException("File with name %s not found", self.file_name)

        self.log.info("Find file with name: %s", file.name)
        context["ti"].xcom_push(key="file_uri", value=file.uri)

        return file.model_dump()


class GenAIGeminiListFilesOperator(GoogleCloudBaseOperator):
    """
    List files uploaded to Gemini Files API.

    The Files API lets you store up to 20GB of files per project, with each file not exceeding 2GB in size.
    Files are stored for 48 hours.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "gemini_api_key",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        gemini_api_key: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.gemini_api_key = gemini_api_key

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        files = self.hook.list_files()

        if files:
            xcom_file_names = []
            xcom_files = []
            try:
                for file in files:
                    xcom_file_names.append(file.name)
                    xcom_files.append(file.model_dump())
            except RuntimeError:
                self.log.info("%s files found", len(xcom_files))

            context["ti"].xcom_push(key="file_names", value=xcom_file_names)
            return xcom_files

        self.log.info("No files found")


class GenAIGeminiDeleteFileOperator(GoogleCloudBaseOperator):
    """
    Delete file uploaded to Gemini Files API.

    The Files API lets you store up to 20GB of files per project, with each file not exceeding 2GB in size.
    Files are stored for 48 hours.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param file_name: Required. File name in Gemini Files API to delete.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "file_name",
        "gemini_api_key",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        file_name: str,
        gemini_api_key: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.file_name = file_name
        self.gemini_api_key = gemini_api_key

    def execute(self, context: Context):
        self.hook = GenAIGeminiAPIHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            gemini_api_key=self.gemini_api_key,
        )

        try:
            delete_response = self.hook.delete_file(file_name=self.file_name)
        except ClientError:
            raise AirflowException("File %s not found!", self.file_name)

        self.log.info("File %s was successfully deleted!", self.file_name)

        return delete_response.model_dump()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.cloud import aiplatform

from airflow.providers.google.cloud.hooks.vertex_ai.experiment_service import (
    ExperimentHook,
    ExperimentRunHook,
)

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
TEST_REGION: str = "test-region"
TEST_PROJECT_ID: str = "test-project-id"
TEST_EXPERIMENT_NAME = "test_experiment_name"
TEST_EXPERIMENT_RUN_NAME = "test_experiment_run_name"
TEST_EXPERIMENT_DESCRIPTION = "test_description"
TEST_TARGET_STATE = aiplatform.gapic.Execution.State.COMPLETE
TEST_TENSORBOARD = False
TEST_DELETE_BACKING_TENSORBOARD_RUNS = False

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
EXPERIMENT_SERVICE_STRING = "airflow.providers.google.cloud.hooks.vertex_ai.experiment_service.{}"


class TestExperimentWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = ExperimentHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.init"))
    def test_create_experiment(self, mock_init) -> None:
        self.hook.create_experiment(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            experiment_description=TEST_EXPERIMENT_DESCRIPTION,
            experiment_tensorboard=TEST_TENSORBOARD,
        )
        mock_init.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment=TEST_EXPERIMENT_NAME,
            experiment_description=TEST_EXPERIMENT_DESCRIPTION,
            experiment_tensorboard=TEST_TENSORBOARD,
        )

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.Experiment"))
    def test_delete_experiment(self, mock_experiment) -> None:
        self.hook.delete_experiment(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            delete_backing_tensorboard_runs=TEST_DELETE_BACKING_TENSORBOARD_RUNS,
        )
        mock_experiment.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
        )
        mock_experiment.return_value.delete.assert_called_with(
            delete_backing_tensorboard_runs=TEST_DELETE_BACKING_TENSORBOARD_RUNS
        )


class TestExperimentWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_no_default_project_id
        ):
            self.hook = ExperimentHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.init"))
    def test_create_experiment(self, mock_init) -> None:
        self.hook.create_experiment(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            experiment_description=TEST_EXPERIMENT_DESCRIPTION,
            experiment_tensorboard=TEST_TENSORBOARD,
        )
        mock_init.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment=TEST_EXPERIMENT_NAME,
            experiment_description=TEST_EXPERIMENT_DESCRIPTION,
            experiment_tensorboard=TEST_TENSORBOARD,
        )

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.Experiment"))
    def test_delete_experiment(self, mock_experiment) -> None:
        self.hook.delete_experiment(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            delete_backing_tensorboard_runs=TEST_DELETE_BACKING_TENSORBOARD_RUNS,
        )
        mock_experiment.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
        )
        mock_experiment.return_value.delete.assert_called_with(
            delete_backing_tensorboard_runs=TEST_DELETE_BACKING_TENSORBOARD_RUNS
        )


class TestExperimentRunWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = ExperimentRunHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.ExperimentRun"))
    def test_create_experiment_run(self, mock_experiment_run) -> None:
        self.hook.create_experiment_run(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            experiment_run_name=TEST_EXPERIMENT_RUN_NAME,
            experiment_run_tensorboard=TEST_TENSORBOARD,
        )
        mock_experiment_run.create.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment=TEST_EXPERIMENT_NAME,
            run_name=TEST_EXPERIMENT_RUN_NAME,
            state=aiplatform.gapic.Execution.State.NEW,
            tensorboard=TEST_TENSORBOARD,
        )

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.ExperimentRun"))
    def test_delete_experiment_run(self, mock_experiment_run) -> None:
        self.hook.delete_experiment_run(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            experiment_run_name=TEST_EXPERIMENT_RUN_NAME,
        )
        mock_experiment_run.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment=TEST_EXPERIMENT_NAME,
            run_name=TEST_EXPERIMENT_RUN_NAME,
        )
        mock_experiment_run.return_value.delete.assert_called_with(
            delete_backing_tensorboard_run=TEST_DELETE_BACKING_TENSORBOARD_RUNS
        )


class TestExperimentRunWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_no_default_project_id
        ):
            self.hook = ExperimentRunHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.ExperimentRun"))
    def test_create_experiment_run(self, mock_experiment_run) -> None:
        self.hook.create_experiment_run(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            experiment_run_name=TEST_EXPERIMENT_RUN_NAME,
            experiment_run_tensorboard=TEST_TENSORBOARD,
        )
        mock_experiment_run.create.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment=TEST_EXPERIMENT_NAME,
            run_name=TEST_EXPERIMENT_RUN_NAME,
            state=aiplatform.gapic.Execution.State.NEW,
            tensorboard=TEST_TENSORBOARD,
        )

    @mock.patch(EXPERIMENT_SERVICE_STRING.format("aiplatform.ExperimentRun"))
    def test_delete_experiment_run(self, mock_experiment_run) -> None:
        self.hook.delete_experiment_run(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment_name=TEST_EXPERIMENT_NAME,
            experiment_run_name=TEST_EXPERIMENT_RUN_NAME,
        )
        mock_experiment_run.assert_called_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            experiment=TEST_EXPERIMENT_NAME,
            run_name=TEST_EXPERIMENT_RUN_NAME,
        )
        mock_experiment_run.return_value.delete.assert_called_with(
            delete_backing_tensorboard_run=TEST_DELETE_BACKING_TENSORBOARD_RUNS
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Task sensor."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.common.compat.sdk import BaseSensorOperator
from airflow.providers.google.cloud.hooks.tasks import CloudTasksHook
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class TaskQueueEmptySensor(BaseSensorOperator):
    """
    Pulls tasks count from a cloud task queue; waits for queue to return task count as 0.

    :param project_id: the Google Cloud project ID for the subscription (templated)
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param queue_name: The queue name to for which task empty sensing is required.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "location",
        "queue_name",
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        queue_name: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.location = location
        self.project_id = project_id
        self.queue_name = queue_name
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def poke(self, context: Context) -> bool:
        hook = CloudTasksHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        # TODO uncomment page_size once https://issuetracker.google.com/issues/155978649?pli=1 gets fixed
        tasks = hook.list_tasks(
            location=self.location,
            queue_name=self.queue_name,
            # page_size=1
        )

        self.log.info("tasks exhausted in cloud task queue?: %s", (len(tasks) == 0))

        return len(tasks) == 0



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import patch

from google.cloud.language_v1 import (
    AnalyzeEntitiesResponse,
    AnalyzeEntitySentimentResponse,
    AnalyzeSentimentResponse,
    ClassifyTextResponse,
    Document,
)

from airflow.providers.google.cloud.operators.natural_language import (
    CloudNaturalLanguageAnalyzeEntitiesOperator,
    CloudNaturalLanguageAnalyzeEntitySentimentOperator,
    CloudNaturalLanguageAnalyzeSentimentOperator,
    CloudNaturalLanguageClassifyTextOperator,
)

DOCUMENT = Document(
    content="Airflow is a platform to programmatically author, schedule and monitor workflows."
)

CLASSIFY_TEXT_RESPONSE = ClassifyTextResponse()
ANALYZE_ENTITIES_RESPONSE = AnalyzeEntitiesResponse()
ANALYZE_ENTITY_SENTIMENT_RESPONSE = AnalyzeEntitySentimentResponse()
ANALYZE_SENTIMENT_RESPONSE = AnalyzeSentimentResponse()

ENCODING_TYPE = "UTF32"


class TestCloudLanguageAnalyzeEntitiesOperator:
    @patch("airflow.providers.google.cloud.operators.natural_language.CloudNaturalLanguageHook")
    def test_minimal_green_path(self, hook_mock):
        hook_mock.return_value.analyze_entities.return_value = ANALYZE_ENTITIES_RESPONSE
        op = CloudNaturalLanguageAnalyzeEntitiesOperator(task_id="task-id", document=DOCUMENT)
        resp = op.execute({})
        assert resp == {}


class TestCloudLanguageAnalyzeEntitySentimentOperator:
    @patch("airflow.providers.google.cloud.operators.natural_language.CloudNaturalLanguageHook")
    def test_minimal_green_path(self, hook_mock):
        hook_mock.return_value.analyze_entity_sentiment.return_value = ANALYZE_ENTITY_SENTIMENT_RESPONSE
        op = CloudNaturalLanguageAnalyzeEntitySentimentOperator(task_id="task-id", document=DOCUMENT)
        resp = op.execute({})
        assert resp == {}


class TestCloudLanguageAnalyzeSentimentOperator:
    @patch("airflow.providers.google.cloud.operators.natural_language.CloudNaturalLanguageHook")
    def test_minimal_green_path(self, hook_mock):
        hook_mock.return_value.analyze_sentiment.return_value = ANALYZE_SENTIMENT_RESPONSE
        op = CloudNaturalLanguageAnalyzeSentimentOperator(task_id="task-id", document=DOCUMENT)
        resp = op.execute({})
        assert resp == {}


class TestCloudLanguageClassifyTextOperator:
    @patch("airflow.providers.google.cloud.operators.natural_language.CloudNaturalLanguageHook")
    def test_minimal_green_path(self, hook_mock):
        hook_mock.return_value.classify_text.return_value = CLASSIFY_TEXT_RESPONSE
        op = CloudNaturalLanguageClassifyTextOperator(task_id="task-id", document=DOCUMENT)
        resp = op.execute({})
        assert resp == {}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Dataproc links."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

import attr

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.common.compat.sdk import BaseOperatorLink, XCom
from airflow.providers.google.cloud.links.base import BASE_LINK, BaseGoogleLink

if TYPE_CHECKING:
    from airflow.models.taskinstancekey import TaskInstanceKey
    from airflow.providers.common.compat.sdk import Context
    from airflow.providers.google.version_compat import BaseOperator


def __getattr__(name: str) -> Any:
    # PEP-562: deprecate module-level variable
    if name == "DATAPROC_JOB_LOG_LINK":
        # TODO: remove DATAPROC_JOB_LOG_LINK alias in the next major release
        # For backward-compatibility, DATAPROC_JOB_LINK was DATAPROC_JOB_LOG_LINK.
        warnings.warn(
            (
                "DATAPROC_JOB_LOG_LINK has been deprecated and will be removed in the next MAJOR release."
                " Please use DATAPROC_JOB_LINK instead"
            ),
            AirflowProviderDeprecationWarning,
            stacklevel=2,
        )
        return DATAPROC_JOB_LINK
    raise AttributeError(f"module {__name__} has no attribute {name}")


DATAPROC_BASE_LINK = BASE_LINK + "/dataproc"
DATAPROC_JOB_LINK = DATAPROC_BASE_LINK + "/jobs/{job_id}?region={region}&project={project_id}"

DATAPROC_CLUSTER_LINK = (
    DATAPROC_BASE_LINK + "/clusters/{cluster_id}/monitoring?region={region}&project={project_id}"
)
DATAPROC_WORKFLOW_TEMPLATE_LINK = (
    DATAPROC_BASE_LINK + "/workflows/templates/{region}/{workflow_template_id}?project={project_id}"
)
DATAPROC_WORKFLOW_LINK = (
    DATAPROC_BASE_LINK + "/workflows/instances/{region}/{workflow_id}?project={project_id}"
)

DATAPROC_BATCH_LINK = DATAPROC_BASE_LINK + "/batches/{region}/{batch_id}/monitoring?project={project_id}"
DATAPROC_BATCHES_LINK = DATAPROC_BASE_LINK + "/batches?project={project_id}"
DATAPROC_JOB_LINK_DEPRECATED = DATAPROC_BASE_LINK + "/jobs/{resource}?region={region}&project={project_id}"
DATAPROC_CLUSTER_LINK_DEPRECATED = (
    DATAPROC_BASE_LINK + "/clusters/{resource}/monitoring?region={region}&project={project_id}"
)


@attr.s(auto_attribs=True)
class DataprocLink(BaseOperatorLink):
    """
    Helper class for constructing Dataproc resource link.

    .. warning::
       This link is pending to deprecate.
    """

    name = "Dataproc resource"
    key = "conf"

    @staticmethod
    def persist(
        context: Context,
        url: str,
        resource: str,
        region: str,
        project_id: str,
    ):
        context["task_instance"].xcom_push(
            key=DataprocLink.key,
            value={
                "region": region,
                "project_id": project_id,
                "url": url,
                "resource": resource,
            },
        )

    def get_link(
        self,
        operator: BaseOperator,
        *,
        ti_key: TaskInstanceKey,
    ) -> str:
        conf = XCom.get_value(key=self.key, ti_key=ti_key)
        return (
            conf["url"].format(
                region=conf["region"], project_id=conf["project_id"], resource=conf["resource"]
            )
            if conf
            else ""
        )

    def __attrs_post_init__(self):
        # This link is still used into the selected operators
        # - airflow.providers.google.cloud.operators.dataproc.DataprocJobBaseOperator
        # As soon as we remove reference to this link we might deprecate it by add warning message
        # with `stacklevel=3` below in this method.
        ...


@attr.s(auto_attribs=True)
class DataprocListLink(BaseOperatorLink):
    """
    Helper class for constructing list of Dataproc resources link.

    .. warning::
       This link is deprecated.
    """

    name = "Dataproc resources"
    key = "list_conf"

    @staticmethod
    def persist(
        context: Context,
        url: str,
        project_id: str,
    ):
        context["task_instance"].xcom_push(
            key=DataprocListLink.key,
            value={
                "project_id": project_id,
                "url": url,
            },
        )

    def get_link(
        self,
        operator: BaseOperator,
        *,
        ti_key: TaskInstanceKey,
    ) -> str:
        list_conf = XCom.get_value(key=self.key, ti_key=ti_key)
        return (
            list_conf["url"].format(
                project_id=list_conf["project_id"],
            )
            if list_conf
            else ""
        )

    def __attrs_post_init__(self):
        warnings.warn(
            "This DataprocListLink is deprecated.",
            AirflowProviderDeprecationWarning,
            stacklevel=3,
        )


class DataprocClusterLink(BaseGoogleLink):
    """Helper class for constructing Dataproc Cluster Link."""

    name = "Dataproc Cluster"
    key = "dataproc_cluster"
    format_str = DATAPROC_CLUSTER_LINK


class DataprocJobLink(BaseGoogleLink):
    """Helper class for constructing Dataproc Job Link."""

    name = "Dataproc Job"
    key = "dataproc_job"
    format_str = DATAPROC_JOB_LINK


class DataprocWorkflowLink(BaseGoogleLink):
    """Helper class for constructing Dataproc Workflow Link."""

    name = "Dataproc Workflow"
    key = "dataproc_workflow"
    format_str = DATAPROC_WORKFLOW_LINK


class DataprocWorkflowTemplateLink(BaseGoogleLink):
    """Helper class for constructing Dataproc Workflow Template Link."""

    name = "Dataproc Workflow Template"
    key = "dataproc_workflow_template"
    format_str = DATAPROC_WORKFLOW_TEMPLATE_LINK


class DataprocBatchLink(BaseGoogleLink):
    """Helper class for constructing Dataproc Batch Link."""

    name = "Dataproc Batch"
    key = "dataproc_batch"
    format_str = DATAPROC_BATCH_LINK


class DataprocBatchesListLink(BaseGoogleLink):
    """Helper class for constructing Dataproc Batches List Link."""

    name = "Dataproc Batches List"
    key = "dataproc_batches_list"
    format_str = DATAPROC_BATCHES_LINK



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import logging
from typing import Any
from unittest import mock
from unittest.mock import AsyncMock

import pytest
from aiohttp import ClientResponseError, RequestInfo
from gcloud.aio.bigquery import Table
from multidict import CIMultiDict
from yarl import URL

from airflow.providers.google.cloud.hooks.bigquery import BigQueryTableAsyncHook
from airflow.providers.google.cloud.triggers.bigquery import (
    BigQueryCheckTrigger,
    BigQueryGetDataTrigger,
    BigQueryInsertJobTrigger,
    BigQueryIntervalCheckTrigger,
    BigQueryTableExistenceTrigger,
    BigQueryTablePartitionExistenceTrigger,
    BigQueryValueCheckTrigger,
)
from airflow.triggers.base import TriggerEvent

TEST_CONN_ID = "bq_default"
TEST_JOB_ID = "1234"
TEST_GCP_PROJECT_ID = "test-project"
TEST_DATASET_ID = "bq_dataset"
TEST_TABLE_ID = "bq_table"
TEST_LOCATION = "US"
POLLING_PERIOD_SECONDS = 4.0
TEST_SQL_QUERY = "SELECT count(*) from Any"
TEST_PASS_VALUE = 2
TEST_TOLERANCE = 1
TEST_FIRST_JOB_ID = "5678"
TEST_SECOND_JOB_ID = "6789"
TEST_METRIC_THRESHOLDS: dict[str, int] = {}
TEST_DATE_FILTER_COLUMN = "ds"
TEST_DAYS_BACK = -7
TEST_RATIO_FORMULA = "max_over_min"
TEST_IGNORE_ZERO = True
TEST_GCP_CONN_ID = "TEST_GCP_CONN_ID"
TEST_IMPERSONATION_CHAIN = "TEST_SERVICE_ACCOUNT"
TEST_HOOK_PARAMS: dict[str, Any] = {}
TEST_PARTITION_ID = "1234"
TEST_SELECTED_FIELDS = "f0_,f1_"


@pytest.fixture
def insert_job_trigger():
    return BigQueryInsertJobTrigger(
        conn_id=TEST_CONN_ID,
        job_id=TEST_JOB_ID,
        project_id=TEST_GCP_PROJECT_ID,
        dataset_id=TEST_DATASET_ID,
        table_id=TEST_TABLE_ID,
        location=TEST_LOCATION,
        poll_interval=POLLING_PERIOD_SECONDS,
        impersonation_chain=TEST_IMPERSONATION_CHAIN,
    )


@pytest.fixture
def get_data_trigger():
    return BigQueryGetDataTrigger(
        conn_id=TEST_CONN_ID,
        job_id=TEST_JOB_ID,
        project_id=TEST_GCP_PROJECT_ID,
        dataset_id=TEST_DATASET_ID,
        table_id=TEST_TABLE_ID,
        location=None,
        poll_interval=POLLING_PERIOD_SECONDS,
        impersonation_chain=TEST_IMPERSONATION_CHAIN,
        selected_fields=TEST_SELECTED_FIELDS,
    )


@pytest.fixture
def table_existence_trigger():
    return BigQueryTableExistenceTrigger(
        TEST_GCP_PROJECT_ID,
        TEST_DATASET_ID,
        TEST_TABLE_ID,
        TEST_GCP_CONN_ID,
        TEST_HOOK_PARAMS,
        POLLING_PERIOD_SECONDS,
        TEST_IMPERSONATION_CHAIN,
    )


@pytest.fixture
def interval_check_trigger():
    return BigQueryIntervalCheckTrigger(
        conn_id=TEST_CONN_ID,
        first_job_id=TEST_FIRST_JOB_ID,
        second_job_id=TEST_SECOND_JOB_ID,
        project_id=TEST_GCP_PROJECT_ID,
        table=TEST_TABLE_ID,
        metrics_thresholds=TEST_METRIC_THRESHOLDS,
        date_filter_column=TEST_DATE_FILTER_COLUMN,
        days_back=TEST_DAYS_BACK,
        ratio_formula=TEST_RATIO_FORMULA,
        ignore_zero=TEST_IGNORE_ZERO,
        dataset_id=TEST_DATASET_ID,
        table_id=TEST_TABLE_ID,
        poll_interval=POLLING_PERIOD_SECONDS,
        impersonation_chain=TEST_IMPERSONATION_CHAIN,
    )


@pytest.fixture
def check_trigger():
    return BigQueryCheckTrigger(
        conn_id=TEST_CONN_ID,
        job_id=TEST_JOB_ID,
        project_id=TEST_GCP_PROJECT_ID,
        dataset_id=TEST_DATASET_ID,
        table_id=TEST_TABLE_ID,
        location=None,
        poll_interval=POLLING_PERIOD_SECONDS,
        impersonation_chain=TEST_IMPERSONATION_CHAIN,
    )


@pytest.fixture
def value_check_trigger():
    return BigQueryValueCheckTrigger(
        conn_id=TEST_CONN_ID,
        pass_value=TEST_PASS_VALUE,
        job_id=TEST_JOB_ID,
        dataset_id=TEST_DATASET_ID,
        project_id=TEST_GCP_PROJECT_ID,
        sql=TEST_SQL_QUERY,
        table_id=TEST_TABLE_ID,
        tolerance=TEST_TOLERANCE,
        poll_interval=POLLING_PERIOD_SECONDS,
        impersonation_chain=TEST_IMPERSONATION_CHAIN,
    )


class TestBigQueryInsertJobTrigger:
    def test_serialization(self, insert_job_trigger):
        """
        Asserts that the BigQueryInsertJobTrigger correctly serializes its arguments and classpath.
        """
        classpath, kwargs = insert_job_trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.bigquery.BigQueryInsertJobTrigger"
        assert kwargs == {
            "cancel_on_kill": True,
            "conn_id": TEST_CONN_ID,
            "job_id": TEST_JOB_ID,
            "project_id": TEST_GCP_PROJECT_ID,
            "dataset_id": TEST_DATASET_ID,
            "table_id": TEST_TABLE_ID,
            "location": TEST_LOCATION,
            "poll_interval": POLLING_PERIOD_SECONDS,
            "impersonation_chain": TEST_IMPERSONATION_CHAIN,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_bigquery_insert_job_op_trigger_success(self, mock_job_status, insert_job_trigger):
        """
        Tests the BigQueryInsertJobTrigger only fires once the query execution reaches a successful state.
        """
        mock_job_status.return_value = {"status": "success", "message": "Job completed"}

        generator = insert_job_trigger.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent({"status": "success", "message": "Job completed", "job_id": TEST_JOB_ID}) == actual
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook._get_job")
    async def test_bigquery_insert_job_trigger_running(self, mock_get_job, caplog, insert_job_trigger):
        """Test that BigQuery Triggers do not fire while a query is still running."""

        mock_get_job.return_value = mock.MagicMock(state="RUNNING")
        caplog.set_level(logging.INFO)

        task = asyncio.create_task(insert_job_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False

        assert "Bigquery job status is running. Sleeping for 4.0 seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_bigquery_op_trigger_terminated(self, mock_job_status, caplog, insert_job_trigger):
        """Test that BigQuery Triggers fire the correct event in case of an error."""
        mock_job_status.return_value = {
            "status": "error",
            "message": "The conn_id `bq_default` isn't defined",
        }

        generator = insert_job_trigger.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent({"status": "error", "message": "The conn_id `bq_default` isn't defined"}) == actual
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_bigquery_op_trigger_exception(self, mock_job_status, caplog, insert_job_trigger):
        """Test that BigQuery Triggers fire the correct event in case of an error."""
        mock_job_status.side_effect = Exception("Test exception")

        generator = insert_job_trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.cancel_job")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    @mock.patch("airflow.providers.google.cloud.triggers.bigquery.BigQueryInsertJobTrigger.safe_to_cancel")
    async def test_bigquery_insert_job_trigger_cancellation(
        self, mock_get_task_instance, mock_get_job_status, mock_cancel_job, caplog, insert_job_trigger
    ):
        """
        Test that BigQueryInsertJobTrigger handles cancellation correctly, logs the appropriate message,
        and conditionally cancels the job based on the `cancel_on_kill` attribute.
        """
        mock_get_task_instance.return_value = True
        insert_job_trigger.cancel_on_kill = True
        insert_job_trigger.job_id = "1234"

        mock_get_job_status.side_effect = [
            {"status": "running", "message": "Job is still running"},
            asyncio.CancelledError(),
        ]

        mock_cancel_job.return_value = asyncio.Future()
        mock_cancel_job.return_value.set_result(None)

        caplog.set_level(logging.INFO)

        try:
            async for _ in insert_job_trigger.run():
                pass
        except asyncio.CancelledError:
            pass

        assert (
            "Task was killed" in caplog.text
            or "Bigquery job status is running. Sleeping for 4.0 seconds." in caplog.text
        ), "Expected messages about task status or cancellation not found in log."
        mock_cancel_job.assert_awaited_once()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.cancel_job")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    @mock.patch("airflow.providers.google.cloud.triggers.bigquery.BigQueryInsertJobTrigger.safe_to_cancel")
    async def test_bigquery_insert_job_trigger_cancellation_unsafe_cancellation(
        self, mock_safe_to_cancel, mock_get_job_status, mock_cancel_job, caplog, insert_job_trigger
    ):
        """
        Test that BigQueryInsertJobTrigger logs the appropriate message and does not cancel the job
        if safe_to_cancel returns False even when the task is cancelled.
        """
        mock_safe_to_cancel.return_value = False
        insert_job_trigger.cancel_on_kill = True
        insert_job_trigger.job_id = "1234"

        # Simulate the initial job status as running
        mock_get_job_status.side_effect = [
            {"status": "running", "message": "Job is still running"},
            asyncio.CancelledError(),
            {"status": "running", "message": "Job is still running after cancellation"},
        ]

        caplog.set_level(logging.INFO)

        try:
            async for _ in insert_job_trigger.run():
                pass
        except asyncio.CancelledError:
            pass

        assert "Skipping to cancel job" in caplog.text, (
            "Expected message about skipping cancellation not found in log."
        )
        assert mock_get_job_status.call_count == 2, "Job status should be checked multiple times"


class TestBigQueryGetDataTrigger:
    def test_bigquery_get_data_trigger_serialization(self, get_data_trigger):
        """Asserts that the BigQueryGetDataTrigger correctly serializes its arguments and classpath."""

        classpath, kwargs = get_data_trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.bigquery.BigQueryGetDataTrigger"
        assert kwargs == {
            "as_dict": False,
            "conn_id": TEST_CONN_ID,
            "impersonation_chain": TEST_IMPERSONATION_CHAIN,
            "job_id": TEST_JOB_ID,
            "dataset_id": TEST_DATASET_ID,
            "project_id": TEST_GCP_PROJECT_ID,
            "table_id": TEST_TABLE_ID,
            "location": None,
            "selected_fields": TEST_SELECTED_FIELDS,
            "poll_interval": POLLING_PERIOD_SECONDS,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook._get_job")
    async def test_bigquery_get_data_trigger_running(self, mock_get_job, caplog, get_data_trigger):
        """Test that BigQuery Triggers do not fire while a query is still running."""

        mock_get_job.return_value = mock.MagicMock(state="running")
        caplog.set_level(logging.INFO)

        task = asyncio.create_task(get_data_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False

        assert "Bigquery job status is running. Sleeping for 4.0 seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_bigquery_get_data_trigger_terminated(self, mock_job_status, caplog, get_data_trigger):
        """Test that BigQuery Triggers fire the correct event in case of an error."""
        # Set the status to a value other than success or pending

        mock_job_status.return_value = {
            "status": "error",
            "message": "The conn_id `bq_default` isn't defined",
        }

        generator = get_data_trigger.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent({"status": "error", "message": "The conn_id `bq_default` isn't defined"}) == actual
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_bigquery_get_data_trigger_exception(self, mock_job_status, caplog, get_data_trigger):
        """Test that BigQuery Triggers fire the correct event in case of an error."""
        mock_job_status.side_effect = Exception("Test exception")

        generator = get_data_trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_output")
    async def test_bigquery_get_data_trigger_success_with_data(
        self, mock_job_output, mock_job_status, get_data_trigger
    ):
        """
        Tests that BigQueryGetDataTrigger only fires once the query execution reaches a successful state.
        """
        mock_job_status.return_value = {"status": "success", "message": "Job completed"}
        mock_job_output.return_value = {
            "kind": "bigquery#tableDataList",
            "etag": "test_etag",
            "schema": {
                "fields": [
                    {"name": "f0_", "type": "INTEGER", "mode": "NULLABLE"},
                    {"name": "f1_", "type": "STRING", "mode": "NULLABLE"},
                ]
            },
            "jobReference": {
                "projectId": "test-airflow-providers",
                "jobId": "test_jobid",
                "location": "US",
            },
            "totalRows": "10",
            "rows": [{"f": [{"v": "42"}, {"v": "monthy python"}]}, {"f": [{"v": "42"}, {"v": "fishy fish"}]}],
            "totalBytesProcessed": "0",
            "jobComplete": True,
            "cacheHit": False,
        }

        generator = get_data_trigger.run()
        actual = await generator.asend(None)
        # The extracted row will be parsed and formatted to retrieve the value from the
        # structure - 'rows":[{"f":[{"v":"42"},{"v":"monthy python"}]},{"f":[{"v":"42"},{"v":"fishy fish"}]}]

        assert (
            TriggerEvent(
                {
                    "status": "success",
                    "message": "Job completed",
                    "records": [[42, "monthy python"], [42, "fishy fish"]],
                }
            )
            == actual
        )
        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()


class TestBigQueryCheckTrigger:
    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook._get_job")
    async def test_bigquery_check_trigger_running(self, mock_get_job, caplog, check_trigger):
        """Test that BigQuery Triggers do not fire while a query is still running."""

        mock_get_job.return_value = mock.MagicMock(state="running")

        task = asyncio.create_task(check_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False

        assert "Bigquery job status is running. Sleeping for 4.0 seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_bigquery_check_trigger_terminated(self, mock_job_status, caplog, check_trigger):
        """Test that BigQuery Triggers fire the correct event in case of an error."""
        # Set the status to a value other than success or pending

        mock_job_status.return_value = {
            "status": "error",
            "message": "The conn_id `bq_default` isn't defined",
        }

        generator = check_trigger.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent({"status": "error", "message": "The conn_id `bq_default` isn't defined"}) == actual
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_check_trigger_exception(self, mock_job_status, caplog, check_trigger):
        """Test that BigQuery Triggers fire the correct event in case of an error."""
        mock_job_status.side_effect = Exception("Test exception")

        generator = check_trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual

    def test_check_trigger_serialization(self, check_trigger):
        """Asserts that the BigQueryCheckTrigger correctly serializes its arguments and classpath."""

        classpath, kwargs = check_trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.bigquery.BigQueryCheckTrigger"
        assert kwargs == {
            "conn_id": TEST_CONN_ID,
            "impersonation_chain": TEST_IMPERSONATION_CHAIN,
            "job_id": TEST_JOB_ID,
            "dataset_id": TEST_DATASET_ID,
            "project_id": TEST_GCP_PROJECT_ID,
            "table_id": TEST_TABLE_ID,
            "location": None,
            "poll_interval": POLLING_PERIOD_SECONDS,
            "cancel_on_kill": True,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_output")
    async def test_check_trigger_success_with_data(self, mock_job_output, mock_job_status, check_trigger):
        """
        Test the BigQueryCheckTrigger only fires once the query execution reaches a successful state.
        """
        mock_job_status.return_value = {"status": "success", "message": "Job completed"}
        mock_job_output.return_value = {
            "kind": "bigquery#getQueryResultsResponse",
            "etag": "test_etag",
            "schema": {"fields": [{"name": "f0_", "type": "INTEGER", "mode": "NULLABLE"}]},
            "jobReference": {
                "projectId": "test_airflow-providers",
                "jobId": "test_jobid",
                "location": "US",
            },
            "totalRows": "1",
            "rows": [{"f": [{"v": "22"}]}],
            "totalBytesProcessed": "0",
            "jobComplete": True,
            "cacheHit": False,
        }

        generator = check_trigger.run()
        actual = await generator.asend(None)

        assert TriggerEvent({"status": "success", "message": "Job completed", "records": [22]}) == actual

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_output")
    async def test_check_trigger_success_without_data(self, mock_job_output, mock_job_status, check_trigger):
        """
        Tests that BigQueryCheckTrigger sends TriggerEvent when no rows are available in the query result.
        """
        mock_job_status.return_value = {"status": "success", "message": "Job completed"}
        mock_job_output.return_value = {
            "kind": "bigquery#getQueryResultsResponse",
            "etag": "test_etag",
            "schema": {
                "fields": [
                    {"name": "value", "type": "INTEGER", "mode": "NULLABLE"},
                    {"name": "name", "type": "STRING", "mode": "NULLABLE"},
                    {"name": "ds", "type": "DATE", "mode": "NULLABLE"},
                ]
            },
            "jobReference": {
                "projectId": "test_airflow-airflow-providers",
                "jobId": "test_jobid",
                "location": "US",
            },
            "totalRows": "0",
            "totalBytesProcessed": "0",
            "jobComplete": True,
            "cacheHit": False,
        }

        generator = check_trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "success", "message": "Job completed", "records": None}) == actual


class TestBigQueryIntervalCheckTrigger:
    def test_interval_check_trigger_serialization(self, interval_check_trigger):
        """
        Asserts that the BigQueryIntervalCheckTrigger correctly serializes its arguments and classpath.
        """

        classpath, kwargs = interval_check_trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.bigquery.BigQueryIntervalCheckTrigger"
        assert kwargs == {
            "conn_id": TEST_CONN_ID,
            "impersonation_chain": TEST_IMPERSONATION_CHAIN,
            "first_job_id": TEST_FIRST_JOB_ID,
            "second_job_id": TEST_SECOND_JOB_ID,
            "project_id": TEST_GCP_PROJECT_ID,
            "table": TEST_TABLE_ID,
            "location": None,
            "metrics_thresholds": TEST_METRIC_THRESHOLDS,
            "date_filter_column": TEST_DATE_FILTER_COLUMN,
            "days_back": TEST_DAYS_BACK,
            "ratio_formula": TEST_RATIO_FORMULA,
            "ignore_zero": TEST_IGNORE_ZERO,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_output")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_records")
    async def test_interval_check_trigger_success(
        self, mock_get_records, mock_get_job_output, mock_job_status, interval_check_trigger
    ):
        """
        Tests the BigQueryIntervalCheckTrigger only fires once the query execution reaches a successful state.
        """
        mock_get_records.return_value = {}
        mock_job_status.return_value = {"status": "success", "message": "Job completed"}
        mock_get_job_output.return_value = ["0"]

        generator = interval_check_trigger.run()
        actual = await generator.asend(None)
        assert actual == TriggerEvent({"status": "error", "message": "The second SQL query returned None"})

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_interval_check_trigger_pending(self, mock_job_status, caplog, interval_check_trigger):
        """
        Tests that the BigQueryIntervalCheckTrigger do not fire while a query is still running.
        """
        mock_job_status.return_value = {"status": "pending", "message": "Job pending"}
        caplog.set_level(logging.INFO)

        task = asyncio.create_task(interval_check_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False

        assert "Query is still running..." in caplog.text
        assert f"Sleeping for {POLLING_PERIOD_SECONDS} seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_interval_check_trigger_terminated(self, mock_job_status, interval_check_trigger):
        """Tests the BigQueryIntervalCheckTrigger fires the correct event in case of an error."""
        # Set the status to a value other than success or pending
        mock_job_status.return_value = {
            "status": "error",
            "message": "The conn_id `bq_default` isn't defined",
        }

        generator = interval_check_trigger.run()
        actual = await generator.asend(None)

        assert (
            TriggerEvent(
                {"status": "error", "message": "The conn_id `bq_default` isn't defined", "data": None}
            )
            == actual
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_interval_check_trigger_exception(self, mock_job_status, caplog, interval_check_trigger):
        """Tests that the BigQueryIntervalCheckTrigger fires the correct event in case of an error."""
        mock_job_status.side_effect = Exception("Test exception")
        caplog.set_level(logging.DEBUG)

        generator = interval_check_trigger.run()
        actual = await generator.asend(None)

        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual


class TestBigQueryValueCheckTrigger:
    def test_bigquery_value_check_op_trigger_serialization(self, value_check_trigger):
        """Asserts that the BigQueryValueCheckTrigger correctly serializes its arguments and classpath."""

        classpath, kwargs = value_check_trigger.serialize()

        assert classpath == "airflow.providers.google.cloud.triggers.bigquery.BigQueryValueCheckTrigger"
        assert kwargs == {
            "conn_id": TEST_CONN_ID,
            "impersonation_chain": TEST_IMPERSONATION_CHAIN,
            "pass_value": TEST_PASS_VALUE,
            "job_id": TEST_JOB_ID,
            "dataset_id": TEST_DATASET_ID,
            "project_id": TEST_GCP_PROJECT_ID,
            "location": None,
            "sql": TEST_SQL_QUERY,
            "table_id": TEST_TABLE_ID,
            "tolerance": TEST_TOLERANCE,
            "poll_interval": POLLING_PERIOD_SECONDS,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_records")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_output")
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_value_check_op_trigger_success(
        self, mock_job_status, get_job_output, get_records, value_check_trigger
    ):
        """
        Tests BigQueryValueCheckTrigger only fires once the query execution reaches a successful state.
        """
        mock_job_status.return_value = {"status": "success", "message": "Job completed"}
        get_job_output.return_value = {}
        get_records.return_value = [[2], [4]]

        await value_check_trigger.run().__anext__()
        await asyncio.sleep(0.5)

        generator = value_check_trigger.run()
        actual = await generator.asend(None)
        assert actual == TriggerEvent({"status": "success", "message": "Job completed", "records": [4]})

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_value_check_op_trigger_pending(self, mock_job_status, caplog, value_check_trigger):
        """
        Tests BigQueryValueCheckTrigger only fires once the query execution reaches a successful state.
        """
        mock_job_status.return_value = {"status": "pending", "message": "Job pending"}
        caplog.set_level(logging.INFO)

        task = asyncio.create_task(value_check_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was returned
        assert task.done() is False

        assert "Query is still running..." in caplog.text

        assert f"Sleeping for {POLLING_PERIOD_SECONDS} seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_value_check_op_trigger_fail(self, mock_job_status, value_check_trigger):
        """
        Tests BigQueryValueCheckTrigger only fires once the query execution reaches a successful state.
        """
        mock_job_status.return_value = {"status": "error", "message": "dummy"}

        generator = value_check_trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "dummy", "records": None}) == actual

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryAsyncHook.get_job_status")
    async def test_value_check_trigger_exception(self, mock_job_status):
        """Tests the BigQueryValueCheckTrigger does not fire if there is an exception."""
        mock_job_status.side_effect = Exception("Test exception")

        trigger = BigQueryValueCheckTrigger(
            conn_id=TEST_CONN_ID,
            sql=TEST_SQL_QUERY,
            pass_value=TEST_PASS_VALUE,
            tolerance=1,
            job_id=TEST_JOB_ID,
            project_id=TEST_GCP_PROJECT_ID,
        )
        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual


class TestBigQueryTableExistenceTrigger:
    def test_table_existence_trigger_serialization(self, table_existence_trigger):
        """
        Asserts that the BigQueryTableExistenceTrigger correctly serializes its arguments and classpath.
        """

        classpath, kwargs = table_existence_trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.bigquery.BigQueryTableExistenceTrigger"
        assert kwargs == {
            "dataset_id": TEST_DATASET_ID,
            "project_id": TEST_GCP_PROJECT_ID,
            "table_id": TEST_TABLE_ID,
            "gcp_conn_id": TEST_GCP_CONN_ID,
            "impersonation_chain": TEST_IMPERSONATION_CHAIN,
            "poll_interval": POLLING_PERIOD_SECONDS,
            "hook_params": TEST_HOOK_PARAMS,
        }

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.triggers.bigquery.BigQueryTableExistenceTrigger._table_exists"
    )
    async def test_big_query_table_existence_trigger_success(
        self, mock_table_exists, table_existence_trigger
    ):
        """Tests success case BigQueryTableExistenceTrigger"""
        mock_table_exists.return_value = True

        generator = table_existence_trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "success", "message": "success"}) == actual

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.triggers.bigquery.BigQueryTableExistenceTrigger._table_exists"
    )
    async def test_table_existence_trigger_pending(self, mock_table_exists, table_existence_trigger):
        """Test that BigQueryTableExistenceTrigger is in loop till the table exist."""
        mock_table_exists.return_value = False

        task = asyncio.create_task(table_existence_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.triggers.bigquery.BigQueryTableExistenceTrigger._table_exists"
    )
    async def test_table_existence_trigger_exception(self, mock_table_exists, table_existence_trigger):
        """Test BigQueryTableExistenceTrigger throws exception if any error."""
        mock_table_exists.side_effect = AsyncMock(side_effect=Exception("Test exception"))

        generator = table_existence_trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryTableAsyncHook.get_table_client")
    async def test_table_exists(self, mock_get_table_client, table_existence_trigger):
        """Test BigQueryTableExistenceTrigger._table_exists async function with mocked value
        and mocked return value"""
        hook = BigQueryTableAsyncHook()
        mock_get_table_client.return_value = AsyncMock(Table)

        res = await table_existence_trigger._table_exists(
            hook, TEST_DATASET_ID, TEST_TABLE_ID, TEST_GCP_PROJECT_ID
        )
        assert res is True

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryTableAsyncHook.get_table_client")
    async def test_table_exists_exception(self, mock_get_table_client, table_existence_trigger):
        """Test BigQueryTableExistenceTrigger._table_exists async function with exception and return False"""
        hook = BigQueryTableAsyncHook()
        mock_get_table_client.side_effect = ClientResponseError(
            history=(),
            request_info=RequestInfo(
                headers=CIMultiDict(),
                real_url=URL("https://example.com"),
                method="GET",
                url=URL("https://example.com"),
            ),
            status=404,
            message="Not Found",
        )

        res = await table_existence_trigger._table_exists(
            hook, TEST_DATASET_ID, TEST_TABLE_ID, TEST_GCP_PROJECT_ID
        )
        expected_response = False
        assert res == expected_response

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryTableAsyncHook.get_table_client")
    async def test_table_exists_raise_exception(self, mock_get_table_client, table_existence_trigger):
        """Test BigQueryTableExistenceTrigger._table_exists async function with raise exception"""
        hook = BigQueryTableAsyncHook()
        mock_get_table_client.side_effect = ClientResponseError(
            history=(),
            request_info=RequestInfo(
                headers=CIMultiDict(),
                real_url=URL("https://example.com"),
                method="GET",
                url=URL("https://example.com"),
            ),
            status=400,
            message="Not Found",
        )

        with pytest.raises(ClientResponseError):
            await table_existence_trigger._table_exists(
                hook, TEST_DATASET_ID, TEST_TABLE_ID, TEST_GCP_PROJECT_ID
            )


class TestBigQueryTablePartitionExistenceTrigger:
    def test_serialization_successfully(self):
        """
        Asserts that the BigQueryTablePartitionExistenceTrigger correctly serializes its arguments
        and classpath.
        """

        trigger = BigQueryTablePartitionExistenceTrigger(
            dataset_id=TEST_DATASET_ID,
            table_id=TEST_TABLE_ID,
            project_id=TEST_GCP_PROJECT_ID,
            partition_id=TEST_PARTITION_ID,
            poll_interval=POLLING_PERIOD_SECONDS,
            gcp_conn_id=TEST_GCP_CONN_ID,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
            hook_params={},
        )

        classpath, kwargs = trigger.serialize()
        assert (
            classpath
            == "airflow.providers.google.cloud.triggers.bigquery.BigQueryTablePartitionExistenceTrigger"
        )
        assert kwargs == {
            "dataset_id": TEST_DATASET_ID,
            "project_id": TEST_GCP_PROJECT_ID,
            "table_id": TEST_TABLE_ID,
            "partition_id": TEST_PARTITION_ID,
            "gcp_conn_id": TEST_GCP_CONN_ID,
            "impersonation_chain": TEST_IMPERSONATION_CHAIN,
            "poll_interval": POLLING_PERIOD_SECONDS,
            "hook_params": TEST_HOOK_PARAMS,
        }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for Google Cloud Managed Service for Apache Kafka testing Topic operations.

Requirements:
    Operator to create a cluster requires GOOGLE_PROVIDER_NETWORK environmental variable
    that will contain the name of the network that will be used for cluster creation.

    Please, note that if you are running this operator in Google Cloud Composer, this value will be set
    automatically and will not require any additional configuration.
    In other cases, the network in which the cluster will be created should be the same as your machine
    is running in.
"""

from __future__ import annotations

import json
import logging
import os
import random
from datetime import datetime
from typing import Any

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.apache.kafka.operators.consume import ConsumeFromTopicOperator
from airflow.providers.apache.kafka.operators.produce import ProduceToTopicOperator
from airflow.providers.google.cloud.operators.managed_kafka import (
    ManagedKafkaCreateClusterOperator,
    ManagedKafkaCreateTopicOperator,
    ManagedKafkaDeleteClusterOperator,
    ManagedKafkaDeleteConsumerGroupOperator,
    ManagedKafkaGetConsumerGroupOperator,
    ManagedKafkaListConsumerGroupsOperator,
    ManagedKafkaUpdateConsumerGroupOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google.gcp_api_client_helpers import create_airflow_connection

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
IS_COMPOSER = bool(os.environ.get("COMPOSER_ENVIRONMENT", ""))
NETWORK = os.environ.get("GOOGLE_PROVIDER_NETWORK") if not IS_COMPOSER else "default"
DAG_ID = "managed_kafka_consumer_group_operations"
LOCATION = "us-central1"

CLUSTER_ID = f"cluster_{DAG_ID}_{ENV_ID}".replace("_", "-")
CLUSTER_CONF = {
    "gcp_config": {
        "access_config": {
            "network_configs": [
                {"subnet": f"projects/{PROJECT_ID}/regions/{LOCATION}/subnetworks/{NETWORK}"},
            ],
        },
    },
    "capacity_config": {
        "vcpu_count": 3,
        "memory_bytes": 3221225472,
    },
}
TOPIC_ID = f"topic_{DAG_ID}_{ENV_ID}".replace("_", "-")
TOPIC_CONF = {
    "partition_count": 3,
    "replication_factor": 3,
}
CONSUMER_GROUP_ID = f"consumer_group_{DAG_ID}_{ENV_ID}".replace("_", "-")
CONNECTION_ID = f"connection_{DAG_ID}_{ENV_ID}"
PORT = "9092"
BOOTSTRAP_URL = f"bootstrap.{CLUSTER_ID}.{LOCATION}.managedkafka.{PROJECT_ID}.cloud.goog:{PORT}"

log = logging.getLogger(__name__)


def producer():
    """Produce and submit 10 messages"""

    for i in range(10):
        now = datetime.now()
        datetime_string = now.strftime("%Y-%m-%d %H:%M:%S")

        message_data = {"random_id": f"{ENV_ID}_{random.randint(1, 100)}", "date_time": datetime_string}

        yield (
            json.dumps(i),
            json.dumps(message_data),
        )


def consumer(message):
    "Take in consumed messages and print its contents to the logs."

    message_content = json.loads(message.value())
    random_id = message_content["random_id"]
    date_time = message_content["date_time"]
    log.info("id: %s, date_time: %s", random_id, date_time)


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "managed_kafka", "consumer_group"],
) as dag:
    create_cluster = ManagedKafkaCreateClusterOperator(
        task_id="create_cluster",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster=CLUSTER_CONF,
        cluster_id=CLUSTER_ID,
    )

    create_topic = ManagedKafkaCreateTopicOperator(
        task_id="create_topic",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        topic_id=TOPIC_ID,
        topic=TOPIC_CONF,
    )

    @task
    def create_connection(connection_id: str):
        conn_extra = {
            "bootstrap.servers": BOOTSTRAP_URL,
            "security.protocol": "SASL_SSL",
            "sasl.mechanisms": "OAUTHBEARER",
            "group.id": CONSUMER_GROUP_ID,
        }
        conn_extra_json = json.dumps(conn_extra)
        connection: dict[str, Any] = {"conn_type": "kafka", "extra": conn_extra_json}
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf=connection,
            is_composer=IS_COMPOSER,
        )

    create_connection_task = create_connection(connection_id=CONNECTION_ID)

    # [START how_to_cloud_managed_kafka_produce_to_topic_operator]
    produce_to_topic = ProduceToTopicOperator(
        task_id="produce_to_topic",
        kafka_config_id=CONNECTION_ID,
        topic=TOPIC_ID,
        producer_function=producer,
        poll_timeout=10,
    )
    # [END how_to_cloud_managed_kafka_produce_to_topic_operator]

    # [START how_to_cloud_managed_kafka_consume_from_topic_operator]
    consume_from_topic = ConsumeFromTopicOperator(
        task_id="consume_from_topic",
        kafka_config_id=CONNECTION_ID,
        topics=[TOPIC_ID],
        apply_function=consumer,
        poll_timeout=20,
        max_messages=20,
        max_batch_size=20,
    )
    # [END how_to_cloud_managed_kafka_consume_from_topic_operator]

    # [START how_to_cloud_managed_kafka_update_consumer_group_operator]
    update_consumer_group = ManagedKafkaUpdateConsumerGroupOperator(
        task_id="update_consumer_group",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        consumer_group_id=CONSUMER_GROUP_ID,
        consumer_group={
            "topics": {},
        },
        update_mask={"paths": ["topics"]},
    )
    # [END how_to_cloud_managed_kafka_update_consumer_group_operator]

    # [START how_to_cloud_managed_kafka_get_consumer_group_operator]
    get_consumer_group = ManagedKafkaGetConsumerGroupOperator(
        task_id="get_consumer_group",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        consumer_group_id=CONSUMER_GROUP_ID,
    )
    # [END how_to_cloud_managed_kafka_get_consumer_group_operator]

    # [START how_to_cloud_managed_kafka_delete_consumer_group_operator]
    delete_consumer_group = ManagedKafkaDeleteConsumerGroupOperator(
        task_id="delete_consumer_group",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        consumer_group_id=CONSUMER_GROUP_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_cloud_managed_kafka_delete_consumer_group_operator]

    delete_cluster = ManagedKafkaDeleteClusterOperator(
        task_id="delete_cluster",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    # [START how_to_cloud_managed_kafka_list_consumer_group_operator]
    list_consumer_groups = ManagedKafkaListConsumerGroupsOperator(
        task_id="list_consumer_groups",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
    )
    # [END how_to_cloud_managed_kafka_list_consumer_group_operator]

    (
        # TEST SETUP
        create_cluster
        >> create_topic
        >> create_connection_task
        >> produce_to_topic
        >> consume_from_topic
        # TEST BODY
        >> update_consumer_group
        >> get_consumer_group
        >> list_consumer_groups
        >> delete_consumer_group
        # TEST TEARDOWN
        >> delete_cluster
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Azure Data Lake Storage to Google Cloud Storage operator."""

from __future__ import annotations

import os
from collections.abc import Sequence
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from airflow.providers.google.cloud.hooks.gcs import GCSHook, _parse_gcs_url

try:
    from airflow.providers.microsoft.azure.hooks.data_lake import AzureDataLakeHook
    from airflow.providers.microsoft.azure.operators.adls import ADLSListOperator
except ModuleNotFoundError as e:
    from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException

    raise AirflowOptionalProviderFeatureException(e)

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class ADLSToGCSOperator(ADLSListOperator):
    """
    Synchronizes an Azure Data Lake Storage path with a GCS bucket.

    :param src_adls: The Azure Data Lake path to find the objects (templated)
    :param dest_gcs: The Google Cloud Storage bucket and prefix to
        store the objects. (templated)
    :param replace: If true, replaces same-named files in GCS
    :param gzip: Option to compress file for upload
    :param azure_data_lake_conn_id: The connection ID to use when
        connecting to Azure Data Lake Storage.
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param google_impersonation_chain: Optional Google service account to impersonate using
        short-term credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

    **Examples**:
        The following Operator would copy a single file named
        ``hello/world.avro`` from ADLS to the GCS bucket ``mybucket``. Its full
        resulting gcs path will be ``gs://mybucket/hello/world.avro`` ::

            copy_single_file = AdlsToGoogleCloudStorageOperator(
                task_id="copy_single_file",
                src_adls="hello/world.avro",
                dest_gcs="gs://mybucket",
                replace=False,
                azure_data_lake_conn_id="azure_data_lake_default",
                gcp_conn_id="google_cloud_default",
            )

        The following Operator would copy all parquet files from ADLS
        to the GCS bucket ``mybucket``. ::

            copy_all_files = AdlsToGoogleCloudStorageOperator(
                task_id='copy_all_files',
                src_adls='*.parquet',
                dest_gcs='gs://mybucket',
                replace=False,
                azure_data_lake_conn_id='azure_data_lake_default',
                gcp_conn_id='google_cloud_default'
            )

         The following Operator would copy all parquet files from ADLS
         path ``/hello/world``to the GCS bucket ``mybucket``. ::

            copy_world_files = AdlsToGoogleCloudStorageOperator(
                task_id='copy_world_files',
                src_adls='hello/world/*.parquet',
                dest_gcs='gs://mybucket',
                replace=False,
                azure_data_lake_conn_id='azure_data_lake_default',
                gcp_conn_id='google_cloud_default'
            )
    """

    template_fields: Sequence[str] = (
        "src_adls",
        "dest_gcs",
        "google_impersonation_chain",
    )
    ui_color = "#f0eee4"

    def __init__(
        self,
        *,
        src_adls: str,
        dest_gcs: str,
        azure_data_lake_conn_id: str,
        gcp_conn_id: str = "google_cloud_default",
        replace: bool = False,
        gzip: bool = False,
        google_impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(path=src_adls, azure_data_lake_conn_id=azure_data_lake_conn_id, **kwargs)

        self.src_adls = src_adls
        self.dest_gcs = dest_gcs
        self.replace = replace
        self.gcp_conn_id = gcp_conn_id
        self.gzip = gzip
        self.google_impersonation_chain = google_impersonation_chain

    def execute(self, context: Context):
        # use the super to list all files in an Azure Data Lake path
        files = super().execute(context)
        g_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.google_impersonation_chain,
        )

        if not self.replace:
            # if we are not replacing -> list all files in the ADLS path
            # and only keep those files which are present in
            # ADLS and not in Google Cloud Storage
            bucket_name, prefix = _parse_gcs_url(self.dest_gcs)
            existing_files = g_hook.list(bucket_name=bucket_name, prefix=prefix)
            files = list(set(files) - set(existing_files))

        if files:
            hook = AzureDataLakeHook(azure_data_lake_conn_id=self.azure_data_lake_conn_id)

            for obj in files:
                with NamedTemporaryFile(mode="wb", delete=True) as f:
                    hook.download_file(local_path=f.name, remote_path=obj)
                    f.flush()
                    dest_gcs_bucket, dest_gcs_prefix = _parse_gcs_url(self.dest_gcs)
                    dest_path = os.path.join(dest_gcs_prefix, obj)
                    self.log.info("Saving file to %s", dest_path)

                    g_hook.upload(
                        bucket_name=dest_gcs_bucket, object_name=dest_path, filename=f.name, gzip=self.gzip
                    )

            self.log.info("All done, uploaded %d files to GCS", len(files))
        else:
            self.log.info("In sync, no files needed to be uploaded to GCS")

        return files



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

# For no Pydantic environment, we need to skip the tests
pytest.importorskip("google.cloud.aiplatform_v1")

from google.api_core.gapic_v1.method import DEFAULT

from airflow.providers.google.cloud.hooks.vertex_ai.model_service import ModelServiceHook

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
TEST_REGION: str = "test-region"
TEST_PROJECT_ID: str = "test-project-id"
TEST_MODEL = None
TEST_PARENT_MODEL = "projects/test-project-id/locations/test-region/models/test-parent-model"
TEST_MODEL_NAME: str = "test_model_name"
TEST_OUTPUT_CONFIG: dict = {}

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
MODEL_SERVICE_STRING = "airflow.providers.google.cloud.hooks.vertex_ai.model_service.{}"
TEST_VERSION_ALIASES = ["new-alias"]


class TestModelServiceWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = ModelServiceHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_delete_model(self, mock_client) -> None:
        self.hook.delete_model(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model=TEST_MODEL_NAME,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.delete_model.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.model_path.assert_called_once_with(
            TEST_PROJECT_ID,
            TEST_REGION,
            TEST_MODEL_NAME,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_export_model(self, mock_client) -> None:
        self.hook.export_model(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model=TEST_MODEL_NAME,
            output_config=TEST_OUTPUT_CONFIG,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.export_model.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                output_config=TEST_OUTPUT_CONFIG,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.model_path.assert_called_once_with(
            TEST_PROJECT_ID, TEST_REGION, TEST_MODEL_NAME
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_list_models(self, mock_client) -> None:
        self.hook.list_models(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.list_models.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                filter=None,
                page_size=None,
                page_token=None,
                read_mask=None,
                order_by=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_upload_model(self, mock_client) -> None:
        self.hook.upload_model(project_id=TEST_PROJECT_ID, region=TEST_REGION, model=TEST_MODEL)
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.upload_model.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                model=TEST_MODEL,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_upload_model_with_parent_model(self, mock_client) -> None:
        self.hook.upload_model(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model=TEST_MODEL, parent_model=TEST_PARENT_MODEL
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.upload_model.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                model=TEST_MODEL,
                parent_model=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_list_model_versions(self, mock_client) -> None:
        self.hook.list_model_versions(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.list_model_versions.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_delete_model_version(self, mock_client) -> None:
        self.hook.delete_model_version(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.delete_model_version.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_get_model(self, mock_client) -> None:
        self.hook.get_model(project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME)
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.get_model.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_set_version_as_default(self, mock_client) -> None:
        self.hook.set_version_as_default(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.merge_version_aliases.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                version_aliases=["default"],
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_add_version_aliases(self, mock_client) -> None:
        self.hook.add_version_aliases(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model_id=TEST_MODEL_NAME,
            version_aliases=TEST_VERSION_ALIASES,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.merge_version_aliases.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                version_aliases=TEST_VERSION_ALIASES,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_delete_version_aliases(self, mock_client) -> None:
        self.hook.delete_version_aliases(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model_id=TEST_MODEL_NAME,
            version_aliases=TEST_VERSION_ALIASES,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.merge_version_aliases.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                version_aliases=["-" + alias for alias in TEST_VERSION_ALIASES],
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )


class TestModelServiceWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_no_default_project_id
        ):
            self.hook = ModelServiceHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_delete_model(self, mock_client) -> None:
        self.hook.delete_model(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model=TEST_MODEL_NAME,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.delete_model.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.model_path.assert_called_once_with(
            TEST_PROJECT_ID,
            TEST_REGION,
            TEST_MODEL_NAME,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_export_model(self, mock_client) -> None:
        self.hook.export_model(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model=TEST_MODEL_NAME,
            output_config=TEST_OUTPUT_CONFIG,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.export_model.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                output_config=TEST_OUTPUT_CONFIG,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.model_path.assert_called_once_with(
            TEST_PROJECT_ID, TEST_REGION, TEST_MODEL_NAME
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_list_models(self, mock_client) -> None:
        self.hook.list_models(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.list_models.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                filter=None,
                page_size=None,
                page_token=None,
                read_mask=None,
                order_by=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_upload_model(self, mock_client) -> None:
        self.hook.upload_model(project_id=TEST_PROJECT_ID, region=TEST_REGION, model=TEST_MODEL)
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.upload_model.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                model=TEST_MODEL,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_upload_model_with_parent_model(self, mock_client) -> None:
        self.hook.upload_model(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model=TEST_MODEL, parent_model=TEST_PARENT_MODEL
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.upload_model.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                model=TEST_MODEL,
                parent_model=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_list_model_versions(self, mock_client) -> None:
        self.hook.list_model_versions(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.list_model_versions.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_delete_model_version(self, mock_client) -> None:
        self.hook.delete_model_version(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.delete_model_version.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_get_model(self, mock_client) -> None:
        self.hook.get_model(project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME)
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.get_model.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_set_version_as_default(self, mock_client) -> None:
        self.hook.set_version_as_default(
            project_id=TEST_PROJECT_ID, region=TEST_REGION, model_id=TEST_MODEL_NAME
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.merge_version_aliases.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                version_aliases=["default"],
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_add_version_aliases(self, mock_client) -> None:
        self.hook.add_version_aliases(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model_id=TEST_MODEL_NAME,
            version_aliases=TEST_VERSION_ALIASES,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.merge_version_aliases.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                version_aliases=TEST_VERSION_ALIASES,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(MODEL_SERVICE_STRING.format("ModelServiceHook.get_model_service_client"))
    def test_delete_version_aliases(self, mock_client) -> None:
        self.hook.delete_version_aliases(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            model_id=TEST_MODEL_NAME,
            version_aliases=TEST_VERSION_ALIASES,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.merge_version_aliases.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.model_path.return_value,
                version_aliases=["-" + alias for alias in TEST_VERSION_ALIASES],
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from unittest import mock

from airflow.providers.google.cloud.transfers.gdrive_to_gcs import GoogleDriveToGCSOperator

FOLDER_ID = os.environ.get("GCP_GDRIVE_FOLDER_ID", "abcd1234")
DRIVE_ID = os.environ.get("GCP_GDRIVE_DRIVE_ID", "abcd1234")
FILE_NAME = os.environ.get("GCP_GDRIVE_TO_GCS_FILE_NAME", "gdrive_to_gcs_file.txt")
BUCKET = os.environ.get("GCP_GDRIVE_TO_GCS_BUCKET", "gdrive-to-gcs-bucket")
OBJECT = "prefix/test.txt"
GCP_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]


class TestGoogleDriveToGCSOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.gdrive_to_gcs.GCSHook")
    @mock.patch("airflow.providers.google.cloud.transfers.gdrive_to_gcs.GoogleDriveHook")
    def test_execute(self, mock_gdrive_hook, mock_gcs_hook):
        context = {}
        op = GoogleDriveToGCSOperator(
            task_id="test_task",
            folder_id=FOLDER_ID,
            file_name=FILE_NAME,
            drive_id=DRIVE_ID,
            bucket_name=BUCKET,
            object_name=OBJECT,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        meta = {"id": "123xyz"}
        mock_gdrive_hook.return_value.get_file_id.return_value = meta

        op.execute(context)
        mock_gdrive_hook.return_value.get_file_id.assert_called_once_with(
            folder_id=FOLDER_ID, file_name=FILE_NAME, drive_id=DRIVE_ID
        )

        mock_gdrive_hook.return_value.download_file.assert_called_once_with(
            file_id=meta["id"], file_handle=mock.ANY
        )

        mock_gcs_hook.return_value.provide_file_and_upload.assert_called_once_with(
            bucket_name=BUCKET, object_name=OBJECT
        )

        assert op.dry_run() is None
